<?php

use MediaWiki\MediaWikiServices;

/**
 * Cut-down copy of User interface for local-interwiki-database
 * user rights manipulation.
 */
class CentralAuthGroupMembershipProxy {
	/**
	 * @var string
	 */
	private $name;

	/**
	 * @var CentralAuthUser
	 */
	private $mGlobalUser;

	/**
	 * @param CentralAuthUser $user
	 */
	private function __construct( CentralAuthUser $user ) {
		$this->name = $user->getName();
		$this->mGlobalUser = $user;
	}

	/**
	 * @param string $wikiID Unused
	 * @param int $id
	 * @return string|false
	 */
	public static function whoIs( $wikiID, $id ) {
		$user = self::newFromId( $id );
		if ( $user ) {
			return $user->name;
		} else {
			return false;
		}
	}

	/**
	 * @param string $name
	 * @return CentralAuthGroupMembershipProxy|null
	 */
	public static function newFromName( $name ) {
		$userNameUtils = MediaWikiServices::getInstance()->getUserNameUtils();
		$name = $userNameUtils->getCanonical( $name );
		if ( $name === false ) {
			return null;
		}
		$globalUser = CentralAuthUser::getPrimaryInstanceByName( $name );
		return $globalUser->exists() ? new CentralAuthGroupMembershipProxy( $globalUser ) : null;
	}

	/**
	 * @param int $id
	 * @return CentralAuthGroupMembershipProxy|null
	 */
	public static function newFromId( $id ) {
		$globalUser = CentralAuthUser::newPrimaryInstanceFromId( $id );
		return $globalUser ? new CentralAuthGroupMembershipProxy( $globalUser ) : null;
	}

	/**
	 * @return int
	 */
	public function getId() {
		return $this->mGlobalUser->getId();
	}

	/**
	 * @return bool
	 */
	public function isAnon() {
		return $this->getId() == 0;
	}

	/**
	 * @return string
	 */
	public function getName() {
		return $this->name;
	}

	/**
	 * @return Title
	 */
	public function getUserPage() {
		return Title::makeTitle( NS_USER, $this->getName() );
	}

	/**
	 * Replaces getUserGroups()
	 * @return mixed
	 */
	public function getGroups() {
		return $this->mGlobalUser->getGlobalGroups();
	}

	/**
	 * Replaces User::getGroupMemberships()
	 * @return UserGroupMembership[] Associative array of (group name => UserGroupMembership object)
	 */
	public function getGroupMemberships() {
		$groups = $this->getGroups();
		return array_combine( $groups, array_map( function ( $group ) {
			return new UserGroupMembership( $this->getId(), $group );
		}, $groups ) );
	}

	/**
	 * replaces addUserGroup
	 * @param string[]|string $group
	 * @param string|null $expiry
	 *
	 * @return bool
	 */
	public function addGroup( $group, $expiry = null ) {
		if ( $expiry !== null ) {
			throw new InvalidArgumentException( __METHOD__ . ' cannot process expiries' );
		}

		$this->mGlobalUser->addToGlobalGroups( $group );
		return true;
	}

	/**
	 * replaces removeUserGroup
	 * @param string $group
	 *
	 * @return bool
	 */
	public function removeGroup( $group ) {
		$this->mGlobalUser->removeFromGlobalGroups( $group );
		return true;
	}

	/**
	 * replaces touchUser
	 */
	public function invalidateCache() {
		$this->mGlobalUser->invalidateCache();
	}

	/**
	 * @param string $wiki
	 * @return bool
	 */
	public function attachedOn( $wiki ) {
		return $this->mGlobalUser->attachedOn( $wiki );
	}
}
